from .query import Query
from .parameter_parser import *
from .classes import Connection, Location, StationBoardEntry

name = "pySBB"


base_url = "http://transport.opendata.ch/v1/"


get_connections = Query(base_url + "connections", Connection, "connections",
                        Parameter("from", optional=False, valid_types=[str]),
                        Parameter("to", optional=False, valid_types=[str]),
                        via=ListParameter("via", Parameter("via", valid_types=[str])),
                        date=DateParameter("date", in_format="%Y-%m-%d", out_format="%Y-%m-%d"),
                        time=DateParameter("time", in_format="%H:%M", out_format="%H:%M"),
                        isArrivalTime=BoolParameter("isArrivalTime"),
                        transportations=ListParameter("transportations", Parameter("transportation", valid_values=["train", "tram", "ship", "bus", "cableway"])),
                        limit=BoundedParameter("limit", lower_bound=1, upper_bound=15, valid_types=[int]),
                        page=BoundedParameter("page", lower_bound=0, upper_bound=3, valid_types=[int]),
                        direct=BoolParameter("direct"),
                        sleeper=BoolParameter("sleeper"),
                        cauchette=BoolParameter("cauchette"),
                        bike=BoolParameter("bike"),
                        accessibility=ListParameter("accessibility", Parameter("accessibility", valid_values=["independent_boarding", "assisted_boarding", "advanced_notice"]))
                        )

get_locations = Query(base_url + "locations", Location, "stations",
                      query=Parameter("query", valid_types=[str]),
                      x=Parameter("x", valid_types=[float]),
                      y=Parameter("y", valid_types=[float]),
                      type=Parameter("type", valid_values=["all", "station", "poi", "address"])
                      )

get_stationboard = Query(base_url + "stationboard", StationBoardEntry, "stationboard",
                         Parameter("station", valid_types=[str]),
                         id=Parameter("id", valid_types=[int]),
                         limit=BoundedParameter("limit", lower_bound=0),
                         transportations=ListParameter("transportations", Parameter("transportation", valid_values=["train", "tram", "ship", "bus", "cableway"])),
                         datetime=DateParameter("datetime", in_format="%Y-%m-%d %H:%M", out_format="%Y-%m-%d %H:%M"),
                         type=Parameter("type", valid_values=["departure", "arrival"])
                         )



#!/usr/bin/python3

# Author: Pascal Schärli
import requests
import json


class Query:
    def __init__(self, url, return_class, return_key, *args, **kwargs):
        self.url = url
        self.args = args
        self.kwargs = kwargs
        self.return_class = return_class
        self.return_key = return_key

    def __call__(self, *args, **kwargs):
        params = {}
        if len(args) != len(self.args):
            raise ValueError("Expected {} arguments but got {}".format(len(self.args), len(args)))

        for parser, value in zip(self.args, args):
            if value is not None:
                params[parser.name] = parser.parse(value)

        for kw, value in kwargs.items():
            if kw not in self.kwargs:
                raise ValueError("Unexpected Keyword argument {}.".format(kw))
            if value is not None:
                parser = self.kwargs[kw]
                params[parser.name] = parser.parse(value)

        response = requests.get(self.url, params=params).json()

        if "errors" in response:
            raise ValueError("The API responded with the following error: " + str(response["errors"]))

        data = response[self.return_key]

        out = []

        for elem in data:
            out.append(self.return_class(elem))

        return out



from datetime import datetime


class Parameter:
    def __init__(self, name, optional=True, default=None, valid_types=None, valid_values=None):
        self.name = name
        self.default = default
        self.valid_types = valid_types
        self.valid_values = valid_values

    def parse(self, value):
        if value is None:
            if self.optional:
                return None
            else:
                raise TypeError("{}: Non-Optional value cannot be None".format(self.name))

        if self.valid_types is not None and type(value) not in self.valid_types:
            expected_types = " or ".join([str(t) for t in self.valid_types])
            raise TypeError("{}: Expected {} but got {} with type {}".format(self.name, expected_types, value, type(value)))
        else:
            processed_value = self.process(value)

            if self.valid_values is not None and processed_value not in self.valid_values:
                expected_values = " or ".join([str(v) for v in self.valid_values])
                raise ValueError("{}: Expected {} but got {}".format(self.name, expected_values, processed_value))
            else:
                return processed_value

    def process(self, value):
        return str(value)


class ListParameter(Parameter):
    def __init__(self, name, element_parameter, *args, min_length=None, max_length=None, **kwargs):
        if element_parameter.valid_types is None:
            super().__init__(name, *args, **kwargs)
        else:
            super().__init__(name, *args, valid_types=[list] + element_parameter.valid_types, **kwargs)
        self.element_parameter = element_parameter
        self.min_length = min_length
        self.max_length = max_length

    def process(self, value):
        if type(value) is not list:
            value = [value]
        out = []
        for elem in value:
            out.append(self.element_parameter.parse(elem))

        if self.min_length is not None and len(out) < self.min_length:
            raise ValueError("{}: List length can't be smaller than {} but is {}".format(self.name, self.min_length, len(out)))
        if self.max_length is not None and len(out) > self.max_length:
            raise ValueError("{}: List length can't be bigger than {} but is {}".format(self.name, self.max_length, len(out)))

        return out


class BoolParameter(Parameter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, valid_types=[bool], **kwargs)

    def process(self, value):
        if value:
            return "1"
        else:
            return "0"


class BoundedParameter(Parameter):
    def __init__(self, *args, lower_bound=None, upper_bound=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def process(self, value):
        if self.lower_bound is not None and value < self.lower_bound:
            raise ValueError("{}: Value can't be smaller than {} but is {}".format(self.name, self.lower_bound, value))
        if self.upper_bound is not None and value > self.upper_bound:
            raise ValueError("{}: Value can't be bigger than {} but is {}".format(self.name, self.upper_bound, value))
        return str(value)


class DateParameter(Parameter):
    def __init__(self, *args, in_format="%Y-%m-%d", out_format="%Y-%m-%d", **kwargs):
        super().__init__(*args, valid_types=[str], **kwargs)
        self.in_format = in_format
        self.out_format = out_format

    def process(self, value):
        datetime_object = datetime.strptime(value, self.in_format)
        return datetime_object.strftime(self.out_format)



# Author: Pascal Schärli

from datetime import timedelta, datetime


class SBBObject():

    # Element names we want to rename
    # (we can't use the variable name from as its a python keyword)
    blacklist = {
        "from": "start",
        "to": "end",
    }

    def __init__(self, data, name):
        self._data = data
        self.__name__ = name

        # Recursively convert dictionary into python classes
        for key, value in data.items():

            # Rename key if its in blacklist
            if key in self.blacklist:
                key = self.blacklist[key]

            # By adding variables to __dict__ we can access them just like
            # Other class variables
            self.__dict__[key] = self._get_object(key, value)

    def __repr__(self):
        return "<{} at 0x{:0x}>".format(self.__name__, id(self))

    def __str__(self):
        """String representation for SBBObjects. Different representations are used depending on what kind of class it is."""

        #
        # Start,Stop points
        #
        if self.__name__ in ["Start", "End", "Stop", "Passlist","Departure","Arrival"]:
            out = "{}".format(self.station)

            info = []
            if self.arrival is not None:
                info.append(self.arrival.strftime("%H:%M"))
            elif self.departure is not None:
                info.append(self.departure.strftime("%H:%M"))
            if self.platform is not None:
                info.append("Plat. {}".format(self.platform))
            if self.delay is not None and self.delay != 0:
                info.append("Delay {} min".format(self.delay))
            if len(info) > 0:
                out = out + " (" + ", ".join(info) + ")"
            return out
        elif self.__name__ in ["Station", "Location"]:
            if self.name is None:
                return ""
            else:
                return self.name
        elif self.__name__ == "Section":
            return "{} -> {}".format(self.departure, self.arrival)
            departure.station
        elif self.__name__ == "Journey":
            return "{}: {}".format(self.name, "->".join([str(p) for p in self.passList]))
        elif self.__name__ == "Coordinate":
            return "({}, {})".format(self.x, self.y)
        else:
            out = []
            for key, value in self.__dict__.items():
                if type(value) in [str, int, float] and not key.startswith("__"):
                    out.append(key + "=" + str(value))
            return "{}: {}".format(self.__name__, ", ".join(out))

    def _get_object(self, key, value):
        if type(value) == dict:
            return SBBObject(value, key.title())
        elif type(value) == list:
            out = []
            for val in value:
                if key[-1] == "s":
                    key = key[:-1]
                out.append(self._get_object(key, val))
            return out
        elif self.isnumber(value):
            if int(value) == float(value):
                return int(value)
            else:
                return float(value)
        elif key in ["departure", "arrival"] and value is not None:
            return datetime.strptime(value[:-5], "%Y-%m-%dT%H:%M:%S")
        elif key == "duration" and value is not None:
            days = int(value[0:2])
            hours = int(value[3:5])
            minutes = int(value[6:8])
            seconds = int(value[9:10])
            return timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)
        else:
            return value

    @staticmethod
    def _timedelta_string(timedelta):
        out = []

        if timedelta.days < 0:
            timedelta = -timedelta
            out.append("-")

        days = timedelta.days
        hours, rem = divmod(timedelta.seconds, 3600)
        minutes, seconds = divmod(rem, 60)

        added = False
        if days != 0:
            out.append("{}d".format(days))
            added = True
        if hours != 0 or added:
            out.append("{}h".format(hours))
            added = True
        if minutes != 0 or added:
            out.append("{}min".format(minutes))
            added = True
        if not added:
            out.append("{}s".format(seconds))
        return " ".join(out)

    @staticmethod
    def isnumber(value):
        try:
            float(value)
            return True
        except:
            return False


class Connection(SBBObject):
    def __init__(self, data):
        super().__init__(data, "Connection")

    def __str__(self):
        return "{} -> {} | {}".format(self.start, self.end, self._timedelta_string(self.duration))


class StationBoardEntry(SBBObject):
    def __init__(self, data):
        super().__init__(data, "StationBoardEntry")

    def __str__(self):
        return "{} -> {} ".format(self.stop, self.end)


class Location(SBBObject):
    def __init__(self, data):
        super().__init__(data, "Location")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pySBB",
    version="1.1.0",
    author="Pascal Schärli",
    author_email="pas.schaerli@sunrise.ch",
    description="An unofficial python wrapper for the SBB api.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pascscha/pySBB",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
